import importlib
import pkgutil

from media_muncher.profile.schemas.base import BaseSchemaGenerator


class SchemaGeneratorFactory:
    _generators = {}

    @classmethod
    def _discover_generators(cls):
        if cls._generators:
            return  # Already discovered

        package = importlib.import_module("media_muncher.profile.schemas")
        prefix = package.__name__ + "."

        for _, module_name, _ in pkgutil.iter_modules(package.__path__, prefix):
            module = importlib.import_module(module_name)
            for attr in dir(module):
                obj = getattr(module, attr)
                if (
                    isinstance(obj, type)
                    and issubclass(obj, BaseSchemaGenerator)
                    and obj is not BaseSchemaGenerator
                ):
                    schema_name = getattr(obj, "schema_name", None)
                    if schema_name:
                        cls._generators[schema_name] = obj

    @classmethod
    def create_generator(
        cls, schema_name: str, config: dict, messages: list
    ) -> BaseSchemaGenerator:
        cls._discover_generators()
        generator_class = cls._generators.get(schema_name)
        if not generator_class:
            raise ValueError(f"Unknown profile schema: {schema_name}")
        return generator_class(config, messages)



from dataclasses import dataclass
from datetime import timedelta
from typing import Optional


@dataclass
class TimelineSpan:
    start: Optional[timedelta] = None
    end: Optional[timedelta] = None
    duration: timedelta = timedelta(0)
    start_trigger: str = "start"
    span_type: str = "content"
    num_segments: int = 0



import math

from media_muncher.framerate import FrameRate


class GopSizeCalculator(object):
    def __init__(
        self,
        video_frame_rate: FrameRate,
        audio_sample_rate: int = None,
        audio_frames_per_packet: int = 1024,
    ):
        # Input data
        self.video_frame_rate: FrameRate = video_frame_rate
        self.audio_sample_rate: int = audio_sample_rate
        self.audio_frames_per_packet: int = audio_frames_per_packet

        self.smallest_recommended_gop_size: int = None
        if self.audio_sample_rate is not None:
            self.smallest_recommended_gop_size = self._calculate_compatible_gop_size()

        # If no audio sample rate is provided, the smallest recommended GOP size is the
        # number of frames in about one second of video
        if self.smallest_recommended_gop_size is None:
            self.smallest_recommended_gop_size = round(self.video_frame_rate.fps)

    def _calculate_compatible_gop_size(self):
        """
        Calculate a GOP size compatible between the video frame rate and audio sample rate.
        """

        # Not possible with ATSC frame rates
        if not self.video_frame_rate.is_integer():
            return None

        vrate = int(float(self.video_frame_rate))
        return (
            vrate
            * self.audio_frames_per_packet
            / math.gcd(vrate * self.audio_frames_per_packet, self.audio_sample_rate)
        )

    def for_target_duration(
        self,
        max_duration: float,
        max_frames: int = None,   # Imposes upper limit on the number of frames in the GOP
        return_max_duration_if_no_compatible_gop_size: bool = False,
    ) -> float:
        # Calculate the nearest duration to the target duration that is compatible
        # between video frame rate and audio sample rate

        if self.smallest_recommended_gop_size is None:
            return None, None

        candidate_durations = []
        candidate_num_frames = []
        dur = 0
        i = 1
        selected_duration = None
        selected_num_frames = None

        while dur < 12:   # max reasonable duration in seconds
            dur = self.smallest_recommended_gop_size * i / self.video_frame_rate.fps
            candidate_durations.append(dur)

            num_frames = self.smallest_recommended_gop_size * i
            candidate_num_frames.append(num_frames)
            
            if max_frames is not None and num_frames > max_frames:
                break

            if dur <= max_duration:
                selected_duration = dur
                selected_num_frames = num_frames
            i += 1

        if return_max_duration_if_no_compatible_gop_size and selected_duration is None:
            return max_duration * self.video_frame_rate.fps, max_duration

        return selected_num_frames, selected_duration



from m3u8 import Playlist
from media_muncher.handlers.hls import HLSHandler
from media_muncher.analysers.hls import HlsAnalyser


class HlsProfileComparer:
    def __init__(self) -> None:
        self.compatibility_map = {}

    def set_target(self, source_handler: HLSHandler):
        self.source_handler = source_handler
        self.source_analyser = HlsAnalyser(source_handler)

    def check_candidate(self, candidate_handler: HLSHandler):
        self.candidate_handler = candidate_handler
        self.candidate_analyser = HlsAnalyser(candidate_handler)

        # Reinitialise the compatibility map
        self.compatibility_map = {}

        # 1. Check renditions
        self._check_renditions()

        # Analyse results
        self._analyse_compatibility()
        return self.compatibility_map

    def _check_renditions(self):
        # TODO - rewrite all to use the Handler directly

        mappings = []
        matched_candidate_renditions = []

        # Check based on source renditions
        for playlist in self.source_handler.document.playlists:
            rendition_map = dict(source=playlist)
            mappings.append(rendition_map)

            source_codecs = self._extract_codecs(playlist)

            # Find all renditions that match the source rendition
            matching_playlists = [
                cand_playlist
                for cand_playlist in self.candidate_handler.document.playlists
                if self._extract_codecs(cand_playlist) == source_codecs
            ]
            matched_candidate_renditions.extend(matching_playlists)
            rendition_map["candidates"] = matching_playlists

        self.compatibility_map["renditions"] = mappings

        # List remaining renditions
        leftover = []
        for playlist in self.candidate_handler.document.playlists:
            if playlist not in matched_candidate_renditions:
                leftover.append(playlist)
        self.compatibility_map["additional_renditions"] = leftover

    def _analyse_compatibility(self):
        is_compatible = True

        if "renditions" in self.compatibility_map:
            for rendition in self.compatibility_map["renditions"]:
                if len(rendition["candidates"]) == 0:
                    is_compatible = False
                    break

        self.compatibility_map["is_compatible"] = is_compatible

    @staticmethod
    def _extract_codecs(playlist: Playlist):
        codecstring = playlist.stream_info.codecs
        return sorted([cs.lower() for cs in codecstring.split(",")])



from media_muncher.format import MediaFormat

from .generic import ContentHandler


class MP4Handler(ContentHandler):
    media_format = MediaFormat.MP4
    content_types = ["video/mp4"]

    def __init__(self, url, content, **kwargs):
        self.url = url
        self._content = content

    def read(self):
        return "Handling MP4 file."

    @staticmethod
    def is_supported_content(content):
        # TODO - add handling by trying to open the video (or at least the start of it)
        return False



from functools import reduce
from math import gcd
from typing import List

ATSC_CLASSIC = {
    29.97: (30000, 1001),
    24.98: (25000, 1001),
    23.976: (24000, 1001),
    59.94: (60000, 1001),
}


class InvalidFrameRateError(ValueError):
    def __init__(self, message):
        super().__init__(message)


class FrameRate(object):
    def __init__(self, value: str | float | int | tuple, second_value: int = None):
        try:
            if second_value is not None and isinstance(value, int):
                self.numerator = value
                self.denominator = second_value

            elif isinstance(value, str):
                self._from_string(value)
            elif isinstance(value, tuple):
                # check that the values are integers
                if not all(isinstance(v, int) for v in value):
                    raise InvalidFrameRateError(
                        f"Invalid input type for FrameRate: {value}"
                    )
                self.numerator, self.denominator = value
            elif isinstance(value, float):
                self._from_float(value)
            elif isinstance(value, int):
                self._from_int(value)

        except ValueError as e:
            raise InvalidFrameRateError(f"Invalid input: {value} - {e}")

        self._convert_to_canonical()

    def _from_float(self, value):
        if int(value) == value:
            self._from_int(value)
            return
        else:
            # If rate is within 0.01 of a known ATSC rate, return it
            for r in ATSC_CLASSIC:
                if abs(value - r) <= 0.011:
                    self.numerator, self.denominator = ATSC_CLASSIC[r]
                    return

        raise InvalidFrameRateError(f"Unrecognised frame rate: {value}")

    def _from_int(self, value: int):
        self.numerator = value
        self.denominator = 1

    def _from_string(self, input: str):
        if "/" in input:
            self.numerator, self.denominator = map(int, input.split("/"))
        else:
            self._from_float(float(input))

    def _convert_to_canonical(self):
        # if denominator can be divided by 1000, divide numerator by 1000
        quotient, remainder = divmod(self.denominator, 1000)
        if remainder == 0:
            self.numerator = int(self.numerator / quotient)
            self.denominator = int(self.denominator / quotient)
            return
        # if 1000 can be divided by denonminator, multiply to get to 1000
        quotient, remainder = divmod(1000, self.denominator)
        if remainder == 0:
            self.numerator = int(self.numerator * quotient)
            self.denominator = int(self.denominator * quotient)
            return

    def __str__(self):
        return f"{self.numerator}/{self.denominator}"

    def __eq__(self, value: "FrameRate"):
        return (
            self.numerator == value.numerator and self.denominator == value.denominator
        )

    def __lt__(self, value: "FrameRate"):
        return self.numerator / self.denominator < value.numerator / value.denominator

    def __gt__(self, value: "FrameRate"):
        return self.numerator / self.denominator > value.numerator / value.denominator

    def __float__(self):
        return self.numerator / self.denominator

    @property
    def fps(self):
        return float(self)

    def to_obj(self, mode="bkt2"):
        if mode == "bkt2":
            return {"num": self.numerator, "den": self.denominator}
        return {"numerator": self.numerator, "denominator": self.denominator}

    def is_integer(self):
        return float(self) == int(float(self))

    @staticmethod
    def get_common_framerate(rates: List["FrameRate"]):
        # Ensure that they all have the same denominator
        denominators = set([r.denominator for r in rates])
        if len(denominators) != 1:
            raise ValueError(
                f"Cannot determine common framerate for different denominators: {denominators}"
            )

        # Ensure that the numerators are all multiples of the lowest one
        numerators = [r.numerator for r in rates]
        if len(numerators) == 1:
            return FrameRate(numerators[0], denominators.pop())
        gcd_num = reduce(gcd, numerators)
        if gcd_num not in numerators:
            raise ValueError(
                f"Cannot determine common framerate for different numerators: {numerators}"
            )
        return FrameRate(gcd_num, denominators.pop())

    def __hash__(self):
        return hash((self.numerator, self.denominator))



import os
import random
from functools import lru_cache
from typing import Callable, Dict, List, Optional
from urllib.parse import parse_qs, urlencode, urlparse, urlunparse

import m3u8
import requests
from loguru import logger

from media_muncher.analysers.hls import HlsAnalyser
from media_muncher.codecstrings import CodecStringParser
from media_muncher.exceptions import MediaHandlerError
from media_muncher.format import MediaFormat
from media_muncher.handlers.generic import ContentHandler


class HLSHandler(ContentHandler):
    media_format = MediaFormat.HLS
    content_types = ["application/x-mpegurl", "application/vnd.apple.mpegurl"]

    def __init__(self, url, content: bytes | None = None, **kwargs):
        super().__init__(url, content, **kwargs)
        self._document: m3u8.M3U8 = None

    @property
    def document(self) -> m3u8.M3U8:
        if not self._document:
            try:
                self._document = m3u8.loads(content=self.content.decode(), uri=self.url)
            except Exception as e:
                raise MediaHandlerError(
                    message="The HLS manifest could not be parsed",
                    original_message=e.args[0],
                )
        return self._document

    def read(self):
        return "Handling HLS file."

    @property
    def inspector(self) -> HlsAnalyser:
        return HlsAnalyser(self)

    @staticmethod
    def is_supported_content(content):
        return content.decode().startswith("#EXTM3U")

    def appears_supported(self) -> bool:
        return self.is_supported_content(self.content)

    def has_children(self) -> bool:
        if self.document.is_variant:
            return True
        return False

    def sub_playlists(self):
        return self.document.playlists + [m for m in self.document.media if m.uri]

    def num_children(self) -> int:
        return len(self.sub_playlists())

    def get_child(self, index: int, additional_query_params: dict = {}):
        playlists = self.sub_playlists()

        child_url = playlists[index - 1].absolute_uri
        if additional_query_params:
            child_url = self._add_query_parameters_from_dict(
                child_url, additional_query_params
            )

        try:
            return HLSHandler(
                url=child_url,
                headers=self.headers,
            )
        except IndexError as e:
            raise MediaHandlerError(
                message=f"The HLS manifest only has {len(self.document.playlists)} renditions.",
                original_message=e.args[0],
            )

    def get_child_by_uri(self, uri: str):
        playlists = self.sub_playlists()
        for i, playlist in enumerate(playlists):
            if playlist.uri == uri:
                return self.get_child(i + 1)
        return None

    @staticmethod
    def _add_query_parameters_from_dict(url: str, new_params: dict):
        parsed_url = urlparse(url)

        # Parse the existing query parameters
        query_params = parse_qs(parsed_url.query)

        # Add the new query parameter
        for key, value in new_params.items():
            query_params[key] = value

        # Reconstruct the query string
        new_query = urlencode(query_params, doseq=True)

        # Reconstruct the full URL with the new query string
        new_url = urlunparse(parsed_url._replace(query=new_query))

        return new_url

    @lru_cache()
    def _fetch_sub(self, uri, cache_buster=None):
        logger.debug(f"Fetching sub-playlist from {uri} with headers {self.headers}")
        try:
            return m3u8.load(
                uri,
                headers=self.headers,
                # TODO - ability to set exact CERT.
                # See https://github.com/globocom/m3u8?tab=readme-ov-file#using-different-http-clients
                verify_ssl=(True if self.verify_ssl is True else False),
            )
        except Exception as e:
            raise MediaHandlerError(
                message=f"The HLS media playlist could not be parsed: {uri}",
                original_message=e.args[0] if e.args and len(e.args) else str(e),
            )

    def protocol_version(self):
        """Returns the protocol version of the HLS

        Returns:
            str
        """
        # Prefer extraction from the sub-playlists
        ver = None
        if self.has_children():
            sub = self._fetch_sub(self.document.playlists[0].absolute_uri)
            ver = sub.version

        if ver is None:
            ver = self.document.version

        if ver is None:
            ver = 3

        return int(ver)

    def is_live(self):
        """Checks if the HLS is a live stream (ie. without an end)

        Returns:
            bool
        """
        # Check the first sub-playlist
        if len(self.document.playlists):
            sub = self._fetch_sub(self.document.playlists[0].absolute_uri)
            if not sub.is_endlist:  # type: ignore
                return True
            else:
                return False

        else:
            return not self.document.is_endlist

    def get_duration(self):
        """Calculates the duration of the stream (in seconds)

        Returns:
            int
        """
        if self.is_live():
            return -1
        else:
            sub = self._fetch_sub(self.document.playlists[0].absolute_uri)
            return sum([seg.duration for seg in sub.segments])

    def num_segments(self):
        """Calculates the number of segments in the stream

        Returns:
            int
        """
        sub = self._fetch_sub(self.document.playlists[0].absolute_uri)
        return len(sub.segments)

    def first_segment_url(self):
        sub = self._fetch_sub(
            self.document.playlists[0].absolute_uri, cache_buster=random.random()
        )
        segment = sub.segments[0]
        return segment.absolute_uri

    def container_format(self):
        """Checks the container format of the segments

        Returns:
            str
        """
        if len(self.document.playlists) == 0:
            raise MediaHandlerError("There seem to be no playlists in this manifest")
        sub = self._fetch_sub(self.document.playlists[0].absolute_uri)

        # We just check if there is a segment map
        if len(sub.segment_map):
            return "ISOBMFF"
        else:
            return "MPEG-TS"

    def has_muxed_audio(self) -> bool:
        """Checks is the audio stream is muxed in with video

        Returns:
            bool
        """
        # TODO - 2 additional use cases:
        #  - video only
        #  - audio only (no video)

        audio_media = [m for m in self.document.media if m.type == "AUDIO"]

        # If there is no media, then must be muxed
        if len(audio_media) == 0:
            return True

        # Otherwise, if the media doesn't have a URI, then must be muxed
        for media in self.document.media:
            if media.uri is None:
                return True
        return False

    def has_audio_only(self) -> bool:
        for playlist in self.document.playlists:
            # extract info from codecs
            cdc = CodecStringParser.parse_multi_codec_string(
                playlist.stream_info.codecs
            )
            # find any rendition without a video codec
            cdc_v = next((d for d in cdc if d.get("type") == "video"), None)
            if not cdc_v:
                return True
        return False

    def target_duration(self):
        if self.has_children():
            sub = self._fetch_sub(self.document.playlists[0].absolute_uri)
            return sub.target_duration
        else:
            return self.document.target_duration

    def standard_segment_duration(self):
        if self.has_children():
            sub = self._fetch_sub(self.document.playlists[0].absolute_uri)
            # Check the duration of all segments
            durations = [seg.duration for seg in sub.segments]

        else:
            durations = [seg.duration for seg in self.document.segments]

        # Crudely, we just pick the duration that is present most often in the playlists
        durations = sorted(durations, key=durations.count, reverse=True)
        return durations[0]

    def get_update_interval(self):
        return self.target_duration()

    def extract_info(self) -> Dict:
        info = {
            "format": "HLS",
            "version": self.protocol_version(),
            "type": "Live" if self.is_live() else "VOD",
            "container": self.container_format(),
            "audio_only": self.has_audio_only(),
            "target_duration": self.target_duration(),
            "duration": (
                "(live)" if self.is_live() else seconds_to_timecode(self.get_duration())
            ),
            "duration (sec)": (
                "(live)" if self.is_live() else f"{self.get_duration():.3f}"
            ),
            "segments": self.num_segments(),
        }

        return info

    def get_segment_for_url(self, url):
        for segment in self.document.segments:
            if segment.uri == url:
                return segment

    def extract_features(self) -> List[Dict]:
        """Extracts essential information from the HLS manifest"""
        arr = []
        index = 0

        if self.document.is_variant:
            for playlist in self.document.playlists:
                index += 1

                si = playlist.stream_info

                data = dict(
                    index=index,
                    type="variant",
                    # uri=playlist.uri,
                    # url=playlist.absolute_uri,
                    codecs=si.codecs,
                )

                # extract info from codecs
                cdc = CodecStringParser.parse_multi_codec_string(si.codecs)
                cdc_v = next((d for d in cdc if d.get("type") == "video"), None)
                cdc_a = next((d for d in cdc if d.get("type") == "audio"), None)

                if cdc_a:
                    data["codeca"] = cdc_a["cc"]
                if cdc_v:
                    data["codecv"] = cdc_v["cc"]
                    data["profilev"] = cdc_v["profile"]
                    data["levelv"] = cdc_v["level"]

                res = (
                    "{} x {}".format(
                        si.resolution[0],
                        si.resolution[1],
                    )
                    if si.resolution
                    else ""
                )
                data["resolution"] = res
                data["bandwidth"] = si.bandwidth

                data["uri_short"] = shorten_url(playlist.uri)

                arr.append(data)

            for media in self.document.media:
                if media.uri:
                    index += 1
                    data = dict(
                        index=index,
                        type="media",
                        uri=media.uri,
                        language=media.language,
                        # url=media.absolute_uri,
                        uri_short=shorten_url(media.uri),
                    )

                    arr.append(data)

        return arr

    def download(
        self,
        output_path: str = None,
        num_segments: int = 0,
        progress_callback: Optional[Callable] = None,
    ):
        sub_mapping = {}

        total_tasks = 1 + self.num_children() * (1 + num_segments)
        logger.debug(f"Number of files to download: {total_tasks}")
        if progress_callback:
            progress_callback(
                f"Number of files to download: {total_tasks}", total=total_tasks
            )

        # Create the output path if it doesn't exist
        if output_path and not os.path.exists(output_path):
            os.makedirs(output_path)

        # Download the sub-playlists first
        for sub_playlist in self.sub_playlists():
            filename = self._download_sub_playlist(
                sub_playlist.uri, output_path, num_segments, progress_callback
            )
            sub_mapping[sub_playlist.uri] = filename

        main_filename = "main.m3u8"
        local_file_path = os.path.join(output_path, main_filename)
        with open(local_file_path, "wb") as f:
            f.write(self.content)

        message = f"Downloaded main playlist to {local_file_path}"
        logger.info(message)
        if progress_callback:
            progress_callback(message)

        return sub_mapping

    def _download_sub_playlist(
        self,
        playlist_uri: str,
        output_path: str,
        num_segments: int = 0,
        progress_callback: Optional[Callable] = None,
    ):
        # extract the relative path from the URI
        # TODO - handle absolute paths
        if playlist_uri.startswith("http"):
            raise NotImplementedError(
                f"The URI {playlist_uri} is an absolute URL, not a relative path"
            )
        if playlist_uri.startswith("../"):
            raise NotImplementedError(
                f"The URI {playlist_uri} is a relative path, not an absolute path"
            )

        relative_path = os.path.dirname(playlist_uri)
        filename = os.path.basename(playlist_uri)

        local_path = os.path.join(output_path, relative_path)
        # Create the output path if it doesn't exist
        if not os.path.exists(local_path):
            os.makedirs(local_path, exist_ok=True)

        sub_handler = self.get_child_by_uri(playlist_uri)

        if num_segments > 0:
            for i in range(0, num_segments):
                # TODO - handle absolute URLs + modify the playlist to point to the local segments
                absolute_uri = sub_handler.document.segments[i].absolute_uri
                segment_uri = sub_handler.document.segments[i].uri
                if segment_uri.startswith("http"):
                    raise NotImplementedError(
                        f"The URI {segment_uri} is an absolute URL, not a relative path"
                    )
                if segment_uri.startswith("../"):
                    raise NotImplementedError(
                        f"The URI {segment_uri} is a relative path, not an absolute path"
                    )

                segment_filename = os.path.basename(segment_uri)
                segment_local_path = os.path.join(local_path, segment_filename)
                with open(segment_local_path, "wb") as f:
                    # make a request (using requests) to the segment URI and write the content to the file
                    # TODO - handle headers and redirects
                    f.write(requests.get(absolute_uri, headers=self.headers).content)

                    message = f"Downloaded segment to {segment_local_path}"
                    logger.info(message)
                    if progress_callback:
                        progress_callback(message)

        local_file_path = os.path.join(local_path, filename)
        with open(local_file_path, "wb") as f:
            f.write(sub_handler.content)

        message = f"Downloaded sub-playlist to {local_file_path}"
        logger.info(message)
        if progress_callback:
            progress_callback(message)

        return filename


def shorten_url(uri):
    u = urlparse(uri)
    shortened_url = u.path[-50:]
    if u.query:
        shortened_url += "?..."
    return shortened_url


def seconds_to_timecode(duration: float, with_milliseconds=False) -> str:
    hours, remainder = divmod(duration, 3600)
    minutes, seconds = divmod(remainder, 60)
    milliseconds = int((seconds - int(seconds)) * 1000)
    seconds = int(seconds)

    if with_milliseconds:
        return f"{int(hours)}:{int(minutes):02d}:{seconds:02d}.{milliseconds:03d}"
    else:
        return f"{int(hours)}:{int(minutes):02d}:{seconds:02d}"



import math
from dataclasses import dataclass


@dataclass
class Resolution:
    width: int
    height: int

    @property
    def aspect_ratio(self) -> float:
        return self.width / self.height

    @property
    def aspect_ratio_str(self) -> str:
        divisor = math.gcd(self.width, self.height)
        return f"{self.width//divisor}:{self.height//divisor}"

    def __str__(self):
        return f"{self.width}x{self.height}"

    def __eq__(self, value: "Resolution"):
        return self.width == value.width and self.height == value.height

    def __lt__(self, value: "Resolution"):
        return self.width * self.height < value.width * value.height

    def __gt__(self, value: "Resolution"):
        return self.width * self.height > value.width * value.height

    def make_even(self):
        return Resolution(_make_even(self.width), _make_even(self.height))


def _make_even(n):
    return n if n % 2 == 0 else n - 1



class Message:
    def __init__(self, level: str, message: str, topic: str = "-"):
        self.level = level
        self.topic = topic
        self.message = message

    def __repr__(self) -> str:
        return f"{self.level.upper()}: {'[' +  self.topic + ']' if self.topic else ''} {self.message}"


class WarningMessage(Message):

    def __init__(self, message: str, topic: str = "-"):
        super().__init__("warning", message, topic)


class ErrorMessage(Message):

    def __init__(self, message: str, topic: str = "-"):
        super().__init__("error", message, topic)


class InfoMessage(Message):

    def __init__(self, message: str, topic: str = "-"):
        super().__init__("info", message, topic)



import re
from datetime import timedelta
from functools import lru_cache
from typing import TYPE_CHECKING, List

from media_muncher.models.timeline_models import TimelineSpan

if TYPE_CHECKING:
    from media_muncher.handlers.hls import HLSHandler

import m3u8
from pymediainfo import MediaInfo

from media_muncher.codecstrings import CodecStringParser
from media_muncher.messages import WarningMessage


class HlsAnalyser:
    def __init__(self, handler: "HLSHandler") -> None:
        self.handler = handler

        self.messages = []

    def extract_renditions(self):
        m3u8_obj: m3u8.M3U8 = self.handler.document

        # Get info about renditions
        audio_renditions = {}
        video_renditions = {}

        for variant in m3u8_obj.playlists:
            resolution = variant.stream_info.resolution
            bandwidth = variant.stream_info.bandwidth
            frame_rate = variant.stream_info.frame_rate

            codecstrings = variant.stream_info.codecs
            codecs = CodecStringParser.parse_multi_codec_string(codecstrings)

            for codec in codecs:
                video_profile = "main"
                if codec["type"] == "video":
                    video_profile = codec.get("profile")
                    codec["resolution"] = resolution or self.get_resolution(
                        variant.absolute_uri
                    )
                    codec["bitrate"] = self._get_bitrate_from_playlist_name(
                        variant.uri, "video", fallback=bandwidth
                    )
                    codec["framerate"] = frame_rate or self.get_framerate(
                        variant.absolute_uri
                    )
                    codec["audio-group"] = variant.stream_info.audio
                    video_renditions[variant.uri] = codec

                if codec["type"] == "audio":
                    fallback_audio_codec = codec
                    # If the variant playlist has no resolution, it's a proper audio playlist
                    if not variant.stream_info.resolution:
                        key = variant.uri
                        codec["bitrate"] = self._get_bitrate_from_playlist_name(
                            variant.uri, "audio", fallback=bandwidth
                        )
                        audio_renditions[key] = codec
                        # TODO - Probably more complex than this...
                    else:
                        # We find the audio playlist through the associate media
                        key = variant.stream_info.audio
                        # TODO - better mechanism to actually extract audio info when muxed in
                        # TODO - support multiple languages and multiple bitrates
                        # We attempt to extract the bitrate from the playlist name
                        codec["bitrate"] = self._get_bitrate_from_playlist_name(
                            variant.uri, "audio"
                        )
                        if codec["bitrate"] is None:
                            # otherwise, we make wild assumptions
                            if video_profile == "baseline":
                                codec["bitrate"] = 64000
                            else:
                                codec["bitrate"] = 128000

                        # extract all audio media
                        audio_media = [m for m in variant.media if m.type == "AUDIO"]

                        if any(m for m in audio_media if m.uri is None):
                            codec["muxed"] = True

                        for m in audio_media:
                            this_codec = codec.copy()
                            this_codec["language"] = m.language
                            audio_renditions[f"{key}-{m.language}"] = this_codec

                # TODO - when audio not muxed in (separate or adjoining streaminf),
                #  adjust bitrate of video to remove audio

                # TODO - extract audio bitrate and sample rate with ffprobe?

        # Still no audio rendition?  That's an old HLS without audio group.
        # We assume there is one, just not named.
        if not audio_renditions:
            audio_renditions["default_audio"] = fallback_audio_codec

        return [*video_renditions.values(), *audio_renditions.values()]

    @staticmethod
    def _get_bitrate_from_playlist_name(playlist_name, media_type, fallback=None):
        # Attempt to extract bitrate from playlist name

        # unified streaming, eg. FBNHD_AC3-audio_384482_eng=384000-video=730000.m3u8
        if media_type == "video":
            match = re.search(r"video=(\d+)", playlist_name)
            if match:
                return int(match.group(1))

        if media_type == "audio":
            match = re.search(r"audio(_[a-z0-9_]+)?=(\d+)", playlist_name)
            if match:
                return int(match.group(2))

        return fallback

    def extract_packaging_info(self):
        return {
            "packaging": "HLS",
            "version": self.handler.protocol_version(),
            "container": self.handler.container_format(),
            "segment_duration_max": self.handler.target_duration(),
            "segment_duration": self.handler.standard_segment_duration(),
            "audio_only": self.handler.has_audio_only(),
            "muxed_audio": self.handler.has_muxed_audio(),
        }

    @lru_cache()
    def _analyse_first_segment(self, playlist_url):
        sub = m3u8.load(playlist_url)
        if sub.segment_map:
            first_segment = sub.segment_map[0]
        else:
            first_segment = sub.segments[0]
        return MediaInfo.parse(first_segment.absolute_uri)

    def get_framerate(self, playlist_url):
        try:
            media_info = self._analyse_first_segment(playlist_url)
            for track in media_info.tracks:
                if track.track_type == "Video":
                    frame_rate = track.frame_rate
                    if not frame_rate:
                        if track.frame_rate_mode == "VFR":
                            self.messages.append(
                                WarningMessage(
                                    f"Variable frame rate found in segments in {playlist_url}"
                                )
                            )
                        else:
                            self.messages.append(
                                WarningMessage(
                                    f"No frame rate found in segments in {playlist_url}"
                                )
                            )
                    else:
                        return float(track.frame_rate)
        except Exception as e:
            self.messages.append(
                WarningMessage(message=f"Unable to analyze media: {e}")
            )
            return None

    def get_resolution(self, playlist_url):
        try:
            media_info = self._analyse_first_segment(playlist_url)
        except Exception as e:
            self.messages.append(WarningMessage(f"Unable to analyze media: {e}"))
            return None
        for track in media_info.tracks:
            if track.track_type == "Video":
                resolution = track.width, track.height
                if not resolution:
                    self.messages.append(
                        WarningMessage(
                            f"No resolution found in segments in {playlist_url}"
                        )
                    )
                else:
                    return resolution

    def get_timeline_spans(self) -> List[TimelineSpan]:
        m3u8_obj: m3u8.M3U8 = self.handler.document

        if self.handler.has_children():
            # pick the first playlist
            handler: HLSHandler = self.handler.get_child(0)
        else:
            handler: HLSHandler = self.handler

        m3u8_obj = handler.document

        spans = []

        def _determine_content_type(segment: m3u8.Segment) -> str:
            if segment.cue_out:
                return "ad"
            else:
                return "content"

        def _determine_trigger(segment: m3u8.Segment) -> str:
            if segment.cue_in:
                return "CUE-IN"
            if segment.cue_out_start:
                return "CUE-OUT"
            if segment.discontinuity:
                return "DISCONTINUITY"
            return ""

        def _starts_new_span(segment: m3u8.Segment) -> bool:
            if segment.discontinuity or segment.cue_in or segment.cue_out_start:
                return True
            return False

        current_span: TimelineSpan = TimelineSpan()

        # find all discontinuities
        segment: m3u8.Segment
        for i, segment in enumerate(m3u8_obj.segments):
            # new span at discontinuities or significant markers
            if _starts_new_span(segment):
                # close the current span
                if current_span.duration > timedelta(0):
                    current_span.end = (
                        segment.current_program_date_time or current_span.duration
                    )
                    spans.append(current_span)

                # and start a new one
                current_span = TimelineSpan()

            # new span need initialising
            if current_span.start is None:
                current_span.start = segment.current_program_date_time or timedelta(
                    seconds=0
                )
                current_span.span_type = _determine_content_type(segment)
                current_span.start_trigger = _determine_trigger(segment)
                current_span.duration = timedelta(seconds=segment.duration)
                current_span.num_segments = 1
            else:
                current_span.duration += timedelta(seconds=segment.duration)
                current_span.num_segments += 1

            # last segment: end of playlist
            if i == len(m3u8_obj.segments) - 1:
                current_span.end = (
                    segment.current_program_date_time or current_span.duration
                ) + timedelta(seconds=segment.duration)
                spans.append(current_span)

        return spans



from fractions import Fraction
from importlib.metadata import version

from media_muncher.messages import InfoMessage, WarningMessage
from media_muncher.profile.schemas.base import BaseSchemaGenerator
from media_muncher.profile.schemas.bkt_v1 import (
    calculate_min_gop_size,
    calculate_recommended_durations,
)


class BktV2ProfileSchemaGenerator(BaseSchemaGenerator):
    schema_name = "bkt-v2"

    def generate(self, renditions, packaging, name: str = ""):
        video_renditions = [r for r in renditions if r["type"] == "video"]
        audio_renditions = [r for r in renditions if r["type"] == "audio"]
        video_ladder = {}
        audio_ladder = {}

        audio_sample_rate = 48000

        target_segment_duration = packaging.get("segment_duration", 4)
        if float(target_segment_duration) == int(target_segment_duration):
            target_segment_duration = int(target_segment_duration)

        if len(video_renditions):
            video_frame_rate = renditions[0]["framerate"]
            if isinstance(video_frame_rate, str):
                video_frame_rate = eval(video_frame_rate)

            if video_frame_rate is None:
                video_frame_rate = self.config.get("framerate")
                self.messages.append(
                    WarningMessage(
                        f"Default video framerate of {video_frame_rate} was selected, since it could not be detected in the source"
                    )
                )

            is_common_frame_rate = all(
                [r["framerate"] == video_frame_rate for r in video_renditions]
            )
            if not is_common_frame_rate:
                self.messages.append(
                    WarningMessage(
                        "The video frame rates are different for different renditions. The output of this tool may not be correct."
                    )
                )
            video_frame_rate_fraction = _to_fraction(video_frame_rate)
            framerate_repr = {
                "num": video_frame_rate_fraction.numerator,
                "den": video_frame_rate_fraction.denominator,
            }

            if video_frame_rate_fraction.denominator > 1:
                self.messages.append(
                    WarningMessage(
                        "Using fractional frame rates is not recommended, as it prevents perfect alignment of audio and video segments"
                    )
                )

            if video_frame_rate_fraction.denominator == 1 and not packaging.get(
                "muxed_audio"
            ):
                min_gop_size = calculate_min_gop_size(
                    video_frame_rate, audio_sample_rate
                )
                (
                    selected_duration,
                    candidate_durations,
                ) = calculate_recommended_durations(
                    video_frame_rate, min_gop_size, target_segment_duration
                )
                if not any(
                    d for d in candidate_durations if d[0] == target_segment_duration
                ):
                    idx = next(
                        (
                            i
                            for i, t in enumerate(candidate_durations)
                            if t[0] == selected_duration[0]
                        ),
                        None,
                    )
                    dur_strings = "\n - " + "\n - ".join(
                        [
                            f'"{s}" (= {g} frames = {d:.3f}s)'
                            for g, d, s in candidate_durations[idx - 1 : idx + 2]
                        ]
                    )
                    self.messages.append(
                        InfoMessage(
                            f"The target segment duration (of {target_segment_duration}s) will cause audio and video segments to not align perfectly, which could cause issues during manifest manipulation.\n"
                            f"For this video frame rate and audio sample rate, it is recommended to create video segments with a GOP with a size multiple of {min_gop_size} frames.\n"
                            f'I have therefore selected "{selected_duration[2]}" (ie. {selected_duration[1]:.3f}s)\n'
                            f"Other nearby segment durations are: {dur_strings}"
                        )
                    )
                    target_segment_duration = selected_duration[2]
            else:
                selected_duration = None

            video_ladder = {
                "common": {
                    "preset": self.config["preset"].upper(),
                    "framerate": framerate_repr,
                }
            }
            if "selected_duration" in locals() and selected_duration:
                video_ladder["common"]["gop_size"] = selected_duration[0]
                video_ladder["common"]["keyint_min"] = selected_duration[0]

            for i, r in enumerate(video_renditions):
                height = _make_even(r["resolution"][1])
                rung = {
                    "_codec_info": f"{r['codec']} {r['profile']} {r['level']}",
                    "codec_string": r["codecstring"],
                    "scale": {"width": -2, "height": height},
                    "bitrate": r["bitrate"],
                }

                video_ladder[f"video_{i}"] = rung

        # Audio rungs
        audio_ladder = {
            "common": {"sampling_rate": 48000, "loudnorm": {"i": -23, "tp": -1}}
        }

        for i, r in enumerate(audio_renditions):
            audio_ladder[f"audio_{i}"] = {
                "_codec_info": f"{r['codec']} {r['mode']}",
                "codec_string": r["codecstring"],
                "bitrate": r["bitrate"],
            }

        # Packaging options

        muxed_audio = packaging.get("muxed_audio")
        container = packaging.get("container")
        if (container == "MPEG-TS" and not muxed_audio) or container == "ISOBMFF":
            self.messages.append(
                WarningMessage(
                    "Since audio renditions are separate from video, you should ensure that your segment size is compatible between audio rate and video frame rate"
                )
            )

        packaging_options = {}
        packaging_options["hls_client_manifest_version"] = (
            packaging.get("version", 3) or 3
        )
        target_segment_duration_fraction = _to_fraction(target_segment_duration)
        packaging_options["hls_minimum_fragment_length"] = {
            "num": target_segment_duration_fraction.numerator,
            "den": target_segment_duration_fraction.denominator,
        }

        advanced_packaging_options = {}
        packaging_options["advanced"] = advanced_packaging_options
        if not packaging.get("audio_only"):
            advanced_packaging_options["--hls.no_audio_only"] = ""

        if packaging.get("container") == "ISOBMFF":
            advanced_packaging_options["--hls.fmp4"] = ""
        else:
            if not packaging.get("muxed_audio"):
                advanced_packaging_options["--hls.no_multiplex"] = ""

        profile = {
            "version": "02.00.00",
            "name": name,
            "type": "OFFLINE_TRANSCODING",
            "audios": audio_ladder,
            "videos": video_ladder,
            "packaging": packaging_options,
            "_generator": "bpkio-python-sdk/" + version("bpkio-python-sdk"),
        }

        return profile


def _make_even(n):
    return n if n % 2 == 0 else n - 1


def _to_fraction(s):
    if isinstance(s, str) and "/" in s:
        num, denom = s.split("/")
        return Fraction(int(float(num)), int(float(denom)))
    else:
        return Fraction(str(s))



from typing import List

from loguru import logger
from mpd_inspector.inspector import MPDInspector
from mpd_inspector.parser.mpd_tags import MPD

from media_muncher.codecstrings import CodecStringParser
from media_muncher.handlers.dash import DASHHandler
from media_muncher.models.timeline_models import TimelineSpan


class DashAnalyser:
    def __init__(self, handler: DASHHandler) -> None:
        self.handler = handler
        self.messages = []

    def extract_renditions(self):
        dash_obj: MPD = self.handler.document

        if len(dash_obj.periods) > 1:
            logger.warning(
                "More than 1 period found. Only the first period will be taken into account"
            )

        period = dash_obj.periods[0]

        audio_renditions = {}
        video_renditions = {}

        for adaptation_set in period.adaptation_sets:
            # Audio
            if (
                adaptation_set.content_type == "audio"
                or adaptation_set.mime_type.startswith("audio")
            ):
                for representation in adaptation_set.representations:
                    codec = CodecStringParser.parse_codec_string(
                        representation.codecs or adaptation_set.codecs
                    )

                    codec["bitrate"] = representation.bandwidth
                    codec["language"] = adaptation_set.lang
                    audio_renditions[representation.id] = codec

            # Video
            if (
                adaptation_set.content_type == "video"
                or adaptation_set.mime_type.startswith("video")
            ):
                for representation in adaptation_set.representations:
                    codec = CodecStringParser.parse_codec_string(
                        representation.codecs or adaptation_set.codecs
                    )
                    codec["resolution"] = (representation.width, representation.height)
                    codec["bitrate"] = representation.bandwidth
                    codec["framerate"] = representation.frame_rate
                    video_renditions[representation.id] = codec

        return [*video_renditions.values(), *audio_renditions.values()]

    def extract_packaging_info(self):
        return {"container": "ISOBMFF"}

    def get_timeline_spans(self) -> List[TimelineSpan]:
        inspector: MPDInspector = self.handler.inspector

        spans = []
        for period in inspector.periods:
            segments = list(
                period.adaptation_sets[0]
                .representations[0]
                .segment_information.segments
            )

            end_time = period.end_time.value if period.end_time else None
            duration = period.duration

            # TODO: find a way to get the timeline based on segments, not period

            span = TimelineSpan(
                start=period.start_time.value,
                end=end_time,
                duration=duration,
                num_segments=len(segments),
                span_type="period",
            )
            spans.append(span)

        return spans



import hashlib
import json
from importlib.metadata import version

from media_muncher.framerate import FrameRate
from media_muncher.gop_size import GopSizeCalculator
from media_muncher.h264_levels import H264LevelValidator
from media_muncher.messages import WarningMessage
from media_muncher.profile.schemas.base import BaseSchemaGenerator
from media_muncher.resolution import Resolution
from media_muncher.segment_size import SegmentSizer

PRESET_MAPPING = {
    "VERYFAST": 3,
    "FAST": 5,
    "MEDIUM": 7,
    "SLOW": 9,
    "SLOWER": 11,
    "VERYSLOW": 13,
    "PLACEBO": 15,
}

# TODO - Improve to take into account the different codecs and channel layouts
STANDARD_BITRATES = {
    "mp4a.40.2": {
        48000: {
            "stereo": [96000, 128000, 160000, 192000, 224000, 256000, 320000],
        }
    },
    "mp4a.40.5": {48000: {"stereo": [16000, 32000, 64000, 80000, 96000, 128000]}},
    "ac-3": {
        48000: {
            "stereo": [
                32000,
                48000,
                64000,
                96000,
                112000,
                128000,
                160000,
                192000,
                224000,
                256000,
                320000,
                384000,
                448000,
                512000,
                576000,
                640000,
            ]
        }
    },
}

DEFAULT_FRAMERATE = 25
DEFAULT_CHANNEL_LAYOUT = "stereo"


class BktV2dot1ProfileSchemaGenerator(BaseSchemaGenerator):
    schema_name = "bkt-v2.1"

    def generate(self, renditions, packaging, name: str = ""):
        self.video_renditions = [r for r in renditions if r["type"] == "video"]
        self.audio_renditions = [r for r in renditions if r["type"] == "audio"]

        # Default framerate (in case of audio-only or if cannot be determined)
        self.common_framerate = FrameRate(DEFAULT_FRAMERATE)

        video_ladder = self._process_video_renditions()

        default_audio_sample_rate = 48000

        self.messages.append(
            WarningMessage(
                "This tool currently only assumes an audio sample rate of 48kHz.",
                topic="audio",
            )
        )

        audio_ladder = self._process_audio_renditions(default_audio_sample_rate)

        packaging_options = self._process_packaging(
            packaging, default_audio_sample_rate
        )

        profile = {
            "version": "02.00.05",
            "type": "OFFLINE_TRANSCODING",
            "audios": audio_ladder,
            "videos": video_ladder,
            "packaging": packaging_options,
            "_generator": "bpkio-python-sdk/" + version("bpkio-python-sdk"),
        }

        # Create a hash of the profile
        hash = hashlib.sha256(json.dumps(profile).encode()).hexdigest()
        profile["name"] = f"bic_{name}_{hash}"

        return profile

    def _process_video_renditions(self):
        perf_level = PRESET_MAPPING.get(self.config["preset"].upper(), 7)

        # Video renditions (if any)
        if len(self.video_renditions) == 0:
            return {}
        else:
            # First pass to set the resolution and framerate
            for r in self.video_renditions:
                # TODO - must be done on the basis of the codec
                # Ensure that the resolutions are even
                r["resolution"] = Resolution(*r["resolution"]).make_even()

                # Ensure that we have a framerate for all video renditions
                if "framerate" not in r:
                    r["framerate"] = FrameRate(self.config.get("framerate"))
                    self.messages.append(
                        WarningMessage(
                            f"Default video framerate of {self.config.get('framerate')} was selected, "
                            f"since it could not be detected in the source",
                            topic="framerate",
                        )
                    )
                else:
                    if r["framerate"] is not None:
                        r["framerate"] = FrameRate(r["framerate"])
                    else:
                        r["framerate"] = FrameRate(DEFAULT_FRAMERATE)

            # Get a common framerate
            all_framerates = [r.get("framerate") for r in self.video_renditions]

            # if they're not all the same, find the common one
            if len(set(all_framerates)) == 1:
                multi_rates = False
                self.common_framerate = all_framerates[0]
            else:
                multi_rates = True
                try:
                    self.common_framerate = FrameRate.get_common_framerate(
                        all_framerates
                    )
                except ValueError:
                    self.common_framerate = all_framerates[0]
                    multi_rates = False
                    self.messages.append(
                        WarningMessage(
                            f"The video frame rates are different and incompatible between the renditions. "
                            f"Forcing the first framerate if {self.common_framerate} on all other renditions",
                            topic="framerate",
                        )
                    )
                    for r in self.video_renditions:
                        r["framerate"] = self.common_framerate

            video_ladder = {
                "common": {
                    "perf_level": perf_level,
                }
            }

            if not multi_rates:
                video_ladder["common"]["framerate"] = {
                    "num": self.common_framerate.numerator,
                    "den": self.common_framerate.denominator,
                }
                video_ladder["common"]["gop_size"] = GopSizeCalculator(
                    video_frame_rate=self.common_framerate
                ).for_target_duration(max_duration=1.1, max_frames=300)[0]

            # Second pass to define the ladder
            for i, r in enumerate(self.video_renditions):
                # Adjust the H264 level
                resol, _, bitrate = H264LevelValidator(r["level"]).adjust(
                    resolution=r.get("resolution"),
                    framerate=r.get("framerate"),
                    bitrate=r.get("bitrate"),
                )
                if resol != r.get("resolution"):
                    self.messages.append(
                        WarningMessage(
                            f"Resolution was adjusted to {resol} (from {r.get('resolution')}) to comply with the H264 level {r['level']}",
                            topic="H264 level",
                        )
                    )
                if bitrate != r.get("bitrate"):
                    self.messages.append(
                        WarningMessage(
                            f"Bitrate was adjusted to {bitrate} bps (from {r.get('bitrate')} bps) to comply with the H264 level {r['level']}",
                            topic="H264 level",
                        )
                    )

                rung = {
                    "_codec_info": f"{r['codec']} {r['profile']} {r['level']}",
                    "codec_string": r["codecstring"],
                    "scale": {"width": -2, "height": resol.height},
                    "bitrate": bitrate,
                }
                if multi_rates:
                    rung["framerate"] = {
                        "num": r["framerate"].numerator,
                        "den": r["framerate"].denominator,
                    }
                    rung["gop_size"] = GopSizeCalculator(
                        video_frame_rate=r["framerate"]
                    ).for_target_duration(max_duration=1.1, max_frames=300)[0]

                video_ladder[f"video_{i}"] = rung

            return video_ladder

    def _process_audio_renditions(self, audio_sample_rate):
        channel_layout = DEFAULT_CHANNEL_LAYOUT

        audio_ladder = {
            "common": {
                "sampling_rate": audio_sample_rate,
                "loudnorm": {"i": -23, "tp": -1},
            }
        }

        # Find distinct codecs and languages
        distinct_languages = list(
            set([r.get("language") for r in self.audio_renditions])
        )
        distinct_codecs = list(
            set([r.get("codecstring") for r in self.audio_renditions])
        )

        if any(c not in STANDARD_BITRATES for c in distinct_codecs):
            self.messages.append(
                WarningMessage(
                    "This tool currently only supports the following audio codecs: "
                    + ", ".join(STANDARD_BITRATES.keys()),
                    topic="audio codec",
                )
            )

        # Find distinct bitrates for each codec
        distinct_codec_bitrates = {}
        for ic in distinct_codecs:
            distinct_codec_bitrates[ic] = list(
                set(
                    [
                        r.get("bitrate")
                        for r in self.audio_renditions
                        if r["codecstring"] == ic
                    ]
                )
            )

        rendition_count = 0
        for ic in distinct_codecs:
            first_rendition = next(
                r for r in self.audio_renditions if r["codecstring"] == ic
            )
            this_codec_info = (
                f"{first_rendition['codec']} {first_rendition.get('mode', '')}"
            )

            for ib in distinct_codec_bitrates[ic]:
                rend_name = f"audio_{rendition_count}"
                rend = {
                    "_codec_info": this_codec_info,
                    "codec_string": ic,
                    "channel_layout": channel_layout,
                }
                # Find nearest standard bitrate
                try:
                    nearest_bitrate = max(
                        filter(
                            lambda x: x <= ib,
                            STANDARD_BITRATES[ic][audio_sample_rate][channel_layout],
                        )
                    )
                except Exception as e:
                    nearest_bitrate = STANDARD_BITRATES[ic][audio_sample_rate][
                        channel_layout
                    ][0]
                    self.messages.append(
                        WarningMessage(
                            f"Invalid bitrate. Bitrate was set to {nearest_bitrate} bps (from {ib} bps) to comply with the AAC codec",
                            topic="AAC codec",
                        )
                    )

                if nearest_bitrate != ib:
                    self.messages.append(
                        WarningMessage(
                            f"Bitrate was adjusted to {nearest_bitrate} bps (from {ib} bps) to comply with the AAC codec",
                            topic="AAC codec",
                        )
                    )

                rend["bitrate"] = nearest_bitrate

                if len(distinct_languages) > 1:
                    rend["advanced"] = {}
                    for il in distinct_languages:
                        if il is not None:
                            rend["advanced"][f"--track_language={il}"] = ""

                audio_ladder[rend_name] = rend
                rendition_count += 1

        return audio_ladder

    def _process_packaging(self, packaging, default_audio_sample_rate):
        # Packaging options
        packaging_options = {}

        # Calculate the segment duration
        target_segment_duration = packaging.get("segment_duration", 4)
        segment_sizer = SegmentSizer(
            framerate=self.common_framerate, samplerate=default_audio_sample_rate
        )
        if packaging.get("muxed_audio") is True:
            segment_sizer.set_target_duration(
                target_segment_duration, ignore_audio=True
            )
        else:
            segment_sizer.set_target_duration(target_segment_duration)

        if segment_sizer.actual_duration != target_segment_duration:
            self.messages.append(
                WarningMessage(
                    f"Target segment duration of {target_segment_duration} seconds was adjusted to "
                    f"{round(segment_sizer.actual_duration, 3)} seconds for compatibility "
                    f"with the framerate and audio sample rate.",
                    topic="Segment duration",
                )
            )

        # TODO - check if the gop size is compatible with the segment duration: a segment duration must be a multiple of the gop size

        # HLS packaging config
        hls_config = {}
        packaging_options["hls"] = hls_config
        hls_config["fragment_length"] = {
            "num": segment_sizer.numerator,
            "den": segment_sizer.denominator,
        }

        # Specific HLS configuration
        if packaging.get("packaging") == "HLS":
            # TODO - not a great default if done alongside DASH...
            hls_config["version"] = packaging.get("version", 3)
            advanced_config = {}

            if packaging.get("container") == "ISOBMFF":
                hls_config["fragmented_mp4"] = True

            if packaging.get("container") == "MPEG-TS":
                if packaging.get("muxed_audio") is False:
                    hls_config["audio_not_multiplex"] = True

                if packaging.get("audio_only") is False:
                    # Only if there is a single audio rendition. It helps with BkYou to ensure there is an audio-only rendition
                    if len(self.audio_renditions) == 1:
                        advanced_config["--hls.no_audio_only"] = ""

            if advanced_config:
                hls_config["advanced"] = advanced_config

        dash_config = {}
        packaging_options["dash"] = dash_config

        dash_config["fragment_length"] = {
            "num": segment_sizer.numerator,
            "den": segment_sizer.denominator,
        }

        # TODO - segment template time or number (for compatibility)

        return packaging_options



from media_muncher.format import MediaFormat

from .generic import ContentHandler


class PNGHandler(ContentHandler):
    media_format = MediaFormat.PNG
    content_types = ["image/png"]

    def __init__(self, url, content, **kwargs):
        self.url = url
        self.content = content

    def read(self):
        return "Handling PNG file."

    @staticmethod
    def is_supported_content(content):
        # TODO - add handling by trying to open the image
        return False



from typing import Dict, List

from lxml import etree
from media_muncher.exceptions import MediaHandlerError
from media_muncher.handlers.vast import VASTHandler

from .xml import XMLHandler


class VMAPHandler(XMLHandler):
    content_types = []
    file_extensions = [".vmap"]

    uri_elements = ["vmap:AdTagURI"]

    def __init__(self, url, content: bytes | None = None, **kwargs):
        super().__init__(url, content, **kwargs)
        self._document: etree._Element = None

    @property
    def document(self) -> etree._Element:
        if self._document is None:
            self._document = etree.fromstring(self.content)
        return self._document

    def read(self):
        return "Handling VMAP file."

    @staticmethod
    def is_supported_content(content):
        try:
            root = etree.fromstring(content)
            if root.tag in [
                "{http://www.iab.net/videosuite/vmap}VMAP",
                "{http://www.iab.net/vmap-1.0}VMAP",
            ]:
                return True
        except etree.XMLSyntaxError:
            pass
        return False

    def has_children(self) -> bool:
        return len(self.extract_features()) > 0

    def get_child(self, index: int, **kwargs) -> "VASTHandler":
        features = self.extract_features()
        try:
            return VASTHandler(
                url=features[index - 1]["AdTagURI"], headers=self.headers
            )
        except IndexError as e:
            raise MediaHandlerError(
                message=f"The VMAP only has {len(features)} ad breaks.",
                original_message=e.args[0],
            )

    def extract_features(self) -> List[Dict]:
        ad_breaks = []

        # Extract the timeOffset, breakId, and AdTagURI
        for ad_break in self.document.xpath("//*[local-name() = 'AdBreak']"):
            nsmap = ad_break.nsmap
            time_offset = ad_break.get("timeOffset")
            break_id = ad_break.get("breakId")
            ad_tag_uri = ad_break.findtext(
                "vmap:AdSource/vmap:AdTagURI", namespaces=nsmap
            ).strip()

            ad_breaks.append(
                {"timeOffset": time_offset, "breakId": break_id, "AdTagURI": ad_tag_uri}
            )

        return ad_breaks



from . import factory
from .dash import DASHHandler
from .generic import ContentHandler
from .hls import HLSHandler
from .jpeg import JPEGHandler
from .mp4 import MP4Handler
from .png import PNGHandler
from .vast import VASTHandler
from .vmap import VMAPHandler
from .xml import XMLHandler



from media_muncher.format import MediaFormat

from .generic import ContentHandler


class JPEGHandler(ContentHandler):
    media_format = MediaFormat.JPEG
    content_types = ["image/jpeg"]
    file_extensions = [".jpeg", ".jpg"]

    def __init__(self, url, content, **kwargs):
        super().__init__(url, content, **kwargs)

    def read(self):
        return "Handling JPEG file."

    @staticmethod
    def is_supported_content(content):
        # TODO - add handling by trying to open the image
        return False
